import 'CoreLibs/animation.lua'
import 'CoreLibs/animator.lua'
import 'CoreLibs/easing.lua'
import 'CoreLibs/graphics.lua'
import 'CoreLibs/object.lua'
import 'CoreLibs/ui.lua'

import 'graphics.lua'
import 'utils.lua'

--- @alias XYPair { x: number, y: number }

--- @alias Base { x: number, y: number }

--- @alias Runner { x: number, y: number, nextBase: Base, prevBase: Base, forcedTo: Base }

--- @alias Fielder { onArrive: fun() | nil, x: number | nil, y: number | nil, target: XYPair | nil, speed: number }

local gfx = playdate.graphics

local SCREEN = {
	W = playdate.display.getWidth(),
	H = playdate.display.getHeight()
}

local CENTER = xy(SCREEN.W / 2, SCREEN.H / 2)

local batCrackSound = playdate.sound.sampleplayer.new("sounds/bat-crack-reverb.wav")
local grassBackground = gfx.image.new("images/game/grass.png") --[[@as PlaydateGraphicsImage]]
local playerFrown = gfx.image.new("images/game/player-frown.png") --[[@as PlaydateGraphicsImage]]

local playerImageBlipper = blipper.new(
	100,
    "images/game/player.png",
    "images/game/player-lowhat.png"
)

local BALL_OFFSCREEN = 999

local danceBounceMs = 500
local danceBounceCount = 4
local fielderDanceAnimator = gfx.animator.new(danceBounceMs, 10, 0, easingHill)--, -1 * danceBounceMs * danceBounceCount)
fielderDanceAnimator.repeatCount = danceBounceCount - 1

local pitchFlyTimeMs = 2500
local PITCH_START_X = 200
local PITCH_START_Y, PITCH_END_Y = 90, 250
local pitchAnimator = gfx.animator.new(pitchFlyTimeMs, PITCH_START_Y, PITCH_END_Y, playdate.easingFunctions.outQuint)

local CRANK_OFFSET_DEG = 90
local batBase = xy(CENTER.x - 34, 215)
local batTip = xy(0, 0)

local TAG_DISTANCE = 20

local ball = {
	x = CENTER.x,
	y = BALL_OFFSCREEN,
	size = 6
}

local BAT_LENGTH = 45

local MODES = {
    batting = {},
    running = {}
}

local currentMode = MODES.batting

local ballAnimatorY = gfx.animator.new(0, BALL_OFFSCREEN, BALL_OFFSCREEN, playdate.easingFunctions.linear)
local ballAnimatorX = gfx.animator.new(0, BALL_OFFSCREEN, BALL_OFFSCREEN, playdate.easingFunctions.linear)
-- TODO? Replace this AND ballSizeAnimator with a ballHeightAnimator
-- ...that might lose some of the magic of both. Compromise available? idk
local ballFloatAnimator = gfx.animator.new(2000, -60, 0, easingHill)
local ballSizeMs = 2000
local ballSizeAnimator = gfx.animator.new(ballSizeMs, 9, 6, easingHill)

local HIT_MULT = 10

local deltaSeconds = 0

local FIRST, SECOND, THIRD, HOME = 1, 2, 3, 4

---@type Base[]
local bases = {
    xy(SCREEN.W * 0.93, SCREEN.H * 0.52),
    xy(SCREEN.W * 0.47, SCREEN.H * 0.19),
    xy(SCREEN.W * 0.03, SCREEN.H * 0.52),
    xy(SCREEN.W * 0.474, SCREEN.H * 0.79)
}

---@type table<Base, Base>
local nextBaseMap = {
    [bases[FIRST]] = bases[SECOND],
    [bases[SECOND]] = bases[THIRD],
    [bases[THIRD]] = bases[HOME]
}

function newFielder(speed)
	return {
		speed = speed
	}
end

---@type table<string, Fielder>
local fielders = {
    first = newFielder(40),
    second = newFielder(40),
    shortstop = newFielder(40),
    third = newFielder(40),
    pitcher = newFielder(30),
    catcher = newFielder(20),
    left = newFielder(40),
    center = newFielder(40),
    right = newFielder(40)
}

local PITCHER_POS = {
	x = SCREEN.W * 0.48,
	y = SCREEN.H * 0.40
}

--- Resets the target positions of all fielders to their defaults (at their field positions).
---@param fromOffTheField boolean If provided, also sets all runners' current position to one centralized location.
function resetFielderPositions(fromOffTheField)
	if fromOffTheField then
		for _,fielder in pairs(fielders) do
			fielder.x = CENTER.x
			fielder.y = SCREEN.H
		end
	end

    fielders.first.target = xy(SCREEN.W - 65, SCREEN.H * 0.48)
    fielders.second.target = xy(SCREEN.W * 0.70, SCREEN.H * 0.30)
    fielders.shortstop.target = xy(SCREEN.W * 0.30, SCREEN.H * 0.30)
    fielders.third.target = xy(SCREEN.W * 0.1, SCREEN.H * 0.48)
    fielders.pitcher.target = xy(PITCHER_POS.x, PITCHER_POS.y)
    fielders.catcher.target = xy(SCREEN.W * 0.475, SCREEN.H * 0.92)
    fielders.left.target = xy(SCREEN.W * -1, SCREEN.H * -0.2)
    fielders.center.target = xy(CENTER.x, SCREEN.H * -0.4)
    fielders.right.target = xy(SCREEN.W * 2, SCREEN.H * fielders.left.target.y)
end

local PLAYER_STARTING_X = bases[HOME].x - 40
local PLAYER_STARTING_Y = bases[HOME].y - 3

--- @type Runner[]
local runners = { }

--- @type Runner[]
local outRunners = { }

local nameI = 1
local runnerNames = {"Barbara", "Steve", "Jerry", "Beatrice"}

---@return Runner
function newRunner()
    local new = {
        x = PLAYER_STARTING_X,
        y = PLAYER_STARTING_Y,
        nextBase = nil,
        prevBase = nil,
        name = runnerNames[nameI]
    }
    nameI = nameI + 1
    runners[#runners + 1] = new
    return new
end

---@type Runner | nil
local batter = newRunner()
batter.nextBase = bases[FIRST]
batter.forcedTo = bases[FIRST]

function throwBall(destX, destY, easingFunc, flyTimeMs, floaty)
    if not flyTimeMs then
        flyTimeMs = distanceBetween(ball.x, ball.y, destX, destY) * 5
    end
    ballSizeAnimator:reset(flyTimeMs)
    ballAnimatorY = gfx.animator.new(flyTimeMs, ball.y, destY, easingFunc)
    ballAnimatorX = gfx.animator.new(flyTimeMs, ball.x, destX, easingFunc)
    if floaty then
        ballFloatAnimator:reset(flyTimeMs)
    end
end

function pitch()
    currentMode = MODES.batting
	ballAnimatorX = gfx.animator.new(0, PITCH_START_X, PITCH_START_X, playdate.easingFunctions.linear)
    ballAnimatorY = pitchAnimator
    pitchAnimator:reset()
end

function playdate.AButtonDown()
	if not batter then
		return
	end
    pitch()
end

function playdate.upButtonDown()
    batBase.y = batBase.y - 1
end

function playdate.downButtonDown()
    batBase.y = batBase.y + 1
end

function playdate.rightButtonDown()
    batBase.x = batBase.x + 1
end

function playdate.leftButtonDown()
    batBase.x = batBase.x - 1
end

local elapsedSec = 0
local crankChange = 0

local BASE_HITBOX = 13
--- Returns the base being touched by the runner at (x,y), or nil, if no base is being touched
function isTouchingBase(x, y)
    for _,base in ipairs(bases) do
        if distanceBetween(x, y, base.x, base.y) < BASE_HITBOX then
            return base
        end
    end

    return nil
end

local BALL_CATCH_HITBOX = 3
function isTouchingBall(x, y)
    local ballDistance = distanceBetween(x, y, ball.x, ball.y)
    return ballDistance < BALL_CATCH_HITBOX
end

function updateForcedTos()
end

local outs = 0
local homeScore = 0
local awayScore = 0

function outRunner(runnerIndex)
	print("You're out, runner" .. runnerIndex .. "!")
	outs = math.min(3, outs + 1)
    outRunners[#outRunners+1] = runners[runnerIndex]
    table.remove(runners, runnerIndex)
    updateForcedTos()
    fielderDanceAnimator:reset()
end

function updateFielders()
    local touchingBaseCache = buildCache(
        function(runner)
            return isTouchingBase(runner.x, runner.y)
        end
    )

    for _,fielder in pairs(fielders) do
        -- TODO: Target unforced runners (or their target bases) for tagging
        -- With new Position-based scheme, fielders are now able to set `fielder.target = runner` to track directly
        if fielder.target ~= nil then
            local x, y, distance = normalizeVector(fielder.x, fielder.y, fielder.target.x, fielder.target.y)

            if distance > 1 then
                fielder.x = fielder.x - (x * fielder.speed * deltaSeconds)
                fielder.y = fielder.y - (y * fielder.speed * deltaSeconds)
            else
                if fielder.onArrive then
                    fielder.onArrive()
                end
                fielder.target = nil
            end
        end

        if currentMode == MODES.running and isTouchingBall(fielder.x, fielder.y) then
            local touchedBase = isTouchingBase(fielder.x, fielder.y)
            for i,runner in pairs(runners) do
                local runnerOnBase = touchingBaseCache.get(runner)
                if touchedBase and runner.forcedTo == touchedBase and touchedBase ~= runnerOnBase then
					print("Force out of runner:")
					printTable(runner)
                    outRunner(i)
                elseif not runnerOnBase then
                    local fielderDistance = distanceBetween(runner.x, runner.y, fielder.x, fielder.y)
                    if fielderDistance < TAG_DISTANCE then
						print("Tagged out!")
                        outRunner(i)
                    end
                end
            end
        end
    end
end

--- Returns true if at least one runner is still moving
---@return boolean
function updateRunners()
    local autoRunSpeed = 20
    --autoRunSpeed = 140
    local nonPlayerRunners = filter(runners, function (runner)
        return runner ~= batter
    end)

    local runnerMoved = false
    for _,runner in pairs(nonPlayerRunners) do
        local _, nearestBaseDistance = getNearestOf(bases, runner.x, runner.y)
        if runner.nextBase then
            local nb = runner.nextBase
            local x, y, distance = normalizeVector(runner.x, runner.y, nb.x, nb.y)

            if distance > 1 then
                local mult = 1
                if crankChange < 0 then
                    mult = -1
                end
                local prevX, prevY = runner.x, runner.y
                -- TODO: Drift toward nearest base?
                local autoRun = nearestBaseDistance > 5 and mult * autoRunSpeed * deltaSeconds or 0
                mult = autoRun + (crankChange / 20)
                runner.x = runner.x - (x * mult)
				runner.y = runner.y - (y * mult)
				runnerMoved = runnerMoved or prevX ~= runner.x or prevY ~= runner.y
            else
                runner.nextBase = nextBaseMap[runner.nextBase]
                runner.forcedTo = nil
            end
        end
    end

    return runnerMoved
end

---@return boolean
function ballIsBeingThrown()
    return false
end

---@return boolean
function throwArrivedBeforeRunner()
    return false
end

function getRunnerTargeting(base)
	for _,runner in pairs(runners) do
		if runner.nextBase == base then
			return runner
		end
	end
	return nil
end

---@return Base[]
function getForcedOutTargets()
	local targets = {}
	for _,base in ipairs(bases) do
		local runnerTargetingBase = getRunnerTargeting(base)
		if runnerTargetingBase then
		    targets[#targets+1] = base
		else
			return targets
		end
	end
	return targets
    -- return { bases[FIRST] }
end

--- Returns the position,distance of the basest closest to the runner furthest from a base
---@return { x: number, y: number } | nil, number | nil
function getBaseOfStrandedRunner()
	local farRunnersBase, farDistance
    for _,runner in pairs(runners) do
    	local base, distance = getNearestOf(bases, runner.x, runner.y, function(base)
			return runner.nextBase == base
    	end)
		if farRunnersBase == nil then
		    farRunnersBase = base
		    farDistance = distance
		elseif farDistance < distance then
			farRunnersBase = base
			farDistance = distance
		end
    end

    return farRunnersBase, farDistance
end

--- Returns x,y of the throw target
---@return number|nil, number|nil
function getNextThrowTarget()
	-- TODO: Handle missed throws, check for fielders at target, etc.
    local targets = getForcedOutTargets()
    if #targets ~= 0 then
		return targets[1].x, targets[1].y
	end

	local baseCloseToStrandedRunner = getBaseOfStrandedRunner()
	if baseCloseToStrandedRunner then
		return baseCloseToStrandedRunner.x, baseCloseToStrandedRunner.y
	end
end

local resetFieldersAfterSeconds = 4
local secondsSinceLastRunnerMove = 0

local pitchAfterSeconds = 4
local secondsSincePitchAllowed = -5

function init()
	playdate.display.setRefreshRate(50)
	gfx.setBackgroundColor(gfx.kColorWhite)
	playdate.setMenuImage(gfx.image.new("images/game/menu-image.png"))
	resetFielderPositions(true)
end

function updateBatting()
	secondsSincePitchAllowed = secondsSincePitchAllowed + deltaSeconds
	if secondsSincePitchAllowed > pitchAfterSeconds then
		pitch()
		secondsSincePitchAllowed = 0
	end
	if ball.y < BALL_OFFSCREEN then
		ball.y = ballAnimatorY:currentValue()
		ball.size = 6
	end

    local batAngle = math.rad(playdate.getCrankPosition() + CRANK_OFFSET_DEG)
    batTip.x = batBase.x + (BAT_LENGTH * math.sin(batAngle))
    batTip.y = batBase.y + (BAT_LENGTH * math.cos(batAngle))

	if acceleratedChange >= 0 and
      pointDirectlyUnderLine(ball.x, ball.y, batBase.x, batBase.y, batTip.x, batTip.y, SCREEN.H) then
      	batCrackSound:play()
        currentMode = MODES.running
        ballAngle = batAngle + math.rad(90)

        local mult = math.abs(acceleratedChange / 15)
        local ballVelX = mult * 10 * math.sin(ballAngle)
        local ballVelY = mult * 5 * math.cos(ballAngle)
        if ballVelY > 0 then
            ballVelX = ballVelX * -1
            ballVelY = ballVelY * -1
        end
        ballDestX = ball.x + (ballVelX * HIT_MULT)
        ballDestY = ball.y + (ballVelY * HIT_MULT)
        throwBall(ballDestX, ballDestY, playdate.easingFunctions.outQuint, 2000)

        fielders.first.target = bases[FIRST]
        batter.nextBase = bases[FIRST]
        batter.forcedTo = bases[FIRST]
        batter = nil -- Demote batter to a mere runner

        local chasingFielder = getNearestOf(fielders, ballDestX, ballDestY)
        chasingFielder.target = { x = ballDestX, y = ballDestY }
        chasingFielder.onArrive = function()
            local targetX, targetY = getNextThrowTarget()
            if targetX ~= nil then
				throwBall(targetX, targetY, playdate.easingFunctions.linear, nil, true)
				chasingFielder.onArrive = nil
            end
        end
    end
end

function updateRunning()
	secondsSincePitchAllowed = 0
	ball.size = ballSizeAnimator:currentValue()
	if updateRunners() then
		secondsSinceLastRunnerMove = 0
	else
		secondsSinceLastRunnerMove = secondsSinceLastRunnerMove + deltaSeconds
		if secondsSinceLastRunnerMove > resetFieldersAfterSeconds then
			throwBall(PITCH_START_X, PITCH_START_Y, playdate.easingFunctions.linear, nil, true)
			resetFielderPositions(false)
			currentMode = MODES.batting
			batter = newRunner()
			batter.nextBase = bases[FIRST]
			batter.forcedTo = bases[FIRST]
		end
	end
end

function updateOutRunners()
	for i, runner in ipairs(outRunners) do
		if runner.x < SCREEN.W + 50 and runner.y < SCREEN.H + 50 then
			runner.x = runner.x + (deltaSeconds * 25)
			runner.y = runner.y + (deltaSeconds * 25)
		else
			table.remove(outRunners, i)
		end
	end
end

function updateGameState()
    deltaSeconds = playdate.getElapsedTime() or 0
    playdate.resetElapsedTime()
    elapsedSec = elapsedSec + deltaSeconds
    crankChange, acceleratedChange = playdate.getCrankChange() --[[@as number, number]]

	ball.x = ballAnimatorX:currentValue()
	ball.y = ballAnimatorY:currentValue() + ballFloatAnimator:currentValue()

    if currentMode == MODES.batting then
		updateBatting()
    elseif currentMode == MODES.running then
		updateRunning()
    end

	updateFielders()
	updateOutRunners()
end

local OUT_BUBBLE_SIZE = 6

function drawScoreboard()
	gfx.setDrawOffset(0, 0)
	local y = SCREEN.H * 0.95
	local x = SCREEN.W * 0.05

    gfx.setLineWidth(1)
    gfx.setColor(gfx.kColorBlack)
	gfx.fillRect(x - 15, y - 44, 73, 55)

    gfx.setColor(gfx.kColorWhite)
	for i=outs,2 do
		gfx.drawCircleAtPoint(x + (i * 2.5 * OUT_BUBBLE_SIZE), y, OUT_BUBBLE_SIZE)
	end
	for i=0,(outs - 1) do
		gfx.fillCircleAtPoint(x + (i * 2.5 * OUT_BUBBLE_SIZE), y, OUT_BUBBLE_SIZE)
	end

	local originalDrawMode = gfx.getImageDrawMode()
	gfx.setImageDrawMode(playdate.graphics.kDrawModeInverted)
	gfx.drawText("Home: " .. homeScore, x - 7, y - 40, 100, 20)
	gfx.drawText("Away: " .. awayScore, x - 7, y - 25, 100, 20)
	gfx.setImageDrawMode(originalDrawMode)
end

function playdate.update()
    updateGameState()
    playdate.graphics.animation.blinker.updateAll()

    gfx.clear()

	if ball.x < BALL_OFFSCREEN then
		-- TODO: Show baserunning minimap when panning?
		local offsetX, offsetY = getDrawOffset(SCREEN.W, SCREEN.H, ball.x, ball.y)
		gfx.setDrawOffset(offsetX, offsetY)
    end

    grassBackground:draw(-400, -240)

    gfx.setColor(gfx.kColorBlack)
    gfx.setLineWidth(2)

    gfx.drawCircleAtPoint(ball.x, ball.y, ball.size)

    local fielderDanceHeight = fielderDanceAnimator:currentValue()
    for _,fielder in pairs(fielders) do
        gfx.fillRect(fielder.x, fielder.y - fielderDanceHeight, 14, 25)
    end

    if currentMode == MODES.batting then
		gfx.setLineWidth(5)
        gfx.drawLine(
            batBase.x, batBase.y,
            batTip.x, batTip.y
        )
    end

    if playdate.isCrankDocked() then -- or (crankChange < 2 and currentMode == MODES.running) then
        playdate.ui.crankIndicator:draw()
    end

    -- TODO? Change blip speed depending on runner speed?
    for _,runner in pairs(runners) do
		-- TODO? Scale sprites down as y increases
        playerImageBlipper:draw(false, runner.x, runner.y)
    end
    for _,runner in pairs(outRunners) do
		playerFrown:draw(runner.x, runner.y)
    end

    drawScoreboard()
end

init()