import 'CoreLibs/animation.lua'
import 'CoreLibs/graphics.lua'

-- number = ((number * 2) - 1)
-- return number * number

-- c = c + 0.0 -- convert to float to prevent integer overflow
-- return c * t / d + b

function easingHill(t, b, c, d)
  c = c + 0.0 -- convert to float to prevent integer overflow
  t = t / d
  t = ((t * 2) - 1)
  t = t * t
  return (c * t) + b
end

---@param x number
---@param y number
---@return XYPair
function xy(x, y)
    return {
    	x = x,
    	y = y
    }
end

--- Returns the normalized vector as two values, plus the distance between the given points.
---@return number, number, number
function normalizeVector(x1, y1, x2, y2)
    local distance, a, b = distanceBetween(x1, y1, x2, y2)
    return a / distance, b / distance, distance
end

---@generic T
---@param array T[]
---@param condition fun(T): boolean
---@return T[]
function filter(array, condition)
    local newArray = {}
    for _,element in pairs(array) do
        if condition(element) then
            newArray[#newArray + 1] = element
        end
    end
    return newArray
end

---@generic TIn
---@generic TOut
---@param array TIn[]
---@param mapper fun(TIn): TOut
---@return TOut[]
function map(array, mapper)
    local newArray = {}
    for _,element in pairs(array) do
        newArray[#newArray + 1] = mapper(element)
    end
    return newArray
end

---@return number, number, number
function distanceBetween(x1, y1, x2, y2)
    local a = x1 - x2
    local b = y1 - y2
    return math.sqrt((a*a) + (b*b)), a, b
end

--- Returns true only if the point is below the given line, within the x bounds of said line, and above the bottomBound
--- @return boolean
function pointDirectlyUnderLine(pointX, pointY, lineX1, lineY1, lineX2, lineY2, bottomBound)
    -- This check currently assumes right-handedness.
    -- I.e. it assumes the ball is to the right of batBaseX
    if pointX < lineX1 or pointX > lineX2 or pointY > bottomBound then
        return false
    end

    local m = (lineY2 - lineY1) / (lineX2 - lineX1)

    -- y = mx + b
    -- b = y1 - (m * x1)
    local b = lineY1 - (m * lineX1)
    local yOnLine = (m * pointX) + b
    local yP = pointY
    local yDelta = yOnLine - yP

    return yDelta <= 0
end

--- Returns the nearest position object from the given point, as well as its distance from that point
---@generic T : {x: number, y: number | nil}
---@param array T[]
---@param x number
---@param y number
---@return T,number|nil
function getNearestOf(array, x, y, extraCondition)
    local nearest, nearestDistance = nil, nil
    for _, element in pairs(array) do
    	if not extraCondition or extraCondition(element) then
			if nearest == nil then
				nearest = element
				nearestDistance = distanceBetween(element.x, element.y, x, y)
			else
				local distance = distanceBetween(element.x, element.y, x, y)
				if distance < nearestDistance then
					nearest = element
					nearestDistance = distance
				end
			end
    	end
    end

    return nearest, nearestDistance
end

local NO_VALUE = {}

function buildCache(fetcher)
    local cacheData = {}
    return {
        cacheDate = cacheData,
        get = function(key)
            if cacheData[key] == NO_VALUE then
                return nil
            end
            if cacheData[key] ~= nil then
                return cacheData[key]
            end
            cacheData[key] = fetcher(key) or NO_VALUE
            return cacheData[key]
        end
    }
end

blipper = {}

--- Build an object that simply "blips" between the given images at the given interval.
--- Expects `playdate.graphics.animation.blinker.updateAll()` to be called on every update.
function blipper.new(msInterval, imagePath1, imagePath2)
	local blinker = playdate.graphics.animation.blinker.new(msInterval, msInterval, true)
	blinker:start()
	return {
		blinker = blinker,
		image1 = playdate.graphics.image.new(imagePath1),
		image2 = playdate.graphics.image.new(imagePath2),
		draw = function(self, disableBlipping, x, y)
			local currentImage = (disableBlipping or self.blinker.on) and self.image2 or self.image1
			currentImage:draw(x, y)
		end
	}
end